(* Title:  Interpret/li-tool.sml
   Author: Walther Neuper 2000
   (c) due to copyright terms

Tools for Lucas_Interpreter
*)

signature LUCAS_INTERPRETER_TOOL =
sig
  datatype ass =
    Associated of Tactic.T * term (*..result*) * Proof.context
  | Ass_Weak of Tactic.T * term (*..result*) * Proof.context
  | Not_Associated;
  val associate: Ctree.state -> Proof.context -> (Tactic.T * term (*..Prog_Tac*)) -> ass
  
  val parent_node: Ctree.ctree -> Pos.pos' -> bool * Pos.pos * Rule_Set.T * Proof.context
  val init_pstate: Rule_Set.T -> Proof.context -> I_Model.T -> MethodC.id ->
    Istate.T * Proof.context * Program.T
  val resume_prog: Pos.pos' -> Ctree.ctree -> 
    (Istate.T * Proof.context) * Program.T

  val check_leaf: string -> Proof.context -> Rule_Set.T -> (Env.T * (term option * term)) -> term -> 
      Program.leaf * term option

  val implicit_take: Program.T -> Env.T -> term option
  val get_simplifier: Calc.T -> Rule_Set.T
  val tac_from_prog: Ctree.state -> term -> Tactic.input
(*from isac_test for Minisubpbl*)
  val arguments_from_model : Proof.context -> MethodC.id -> I_Model.T -> term list

\<^isac_test>\<open>
(**)
\<close>
end 

(* traces the leaves (ie. non-tactical nodes) of Prog found by find_next_step *)   
val LI_trace = Attrib.setup_config_bool \<^binding>\<open>LI_trace\<close> (K false);

(**)
structure LItool(**): LUCAS_INTERPRETER_TOOL(**) =
struct
(**)
open Ctree
open Pos

(* find the formal argument of a tactic in case there is only one (e.g. in simplification) *)
fun implicit_take (Rule.Prog prog) env =
    Option.map (subst_atomic env) (Prog_Tac.get_first_argument prog)
  | implicit_take _ _ = raise ERROR "implicit_take: no match";

(*  *)
val error_msg_1 = "ERROR: the guard is missing (#model in 'type met' added in prep_met)."
fun error_msg_2 ctxt d itms =
  ("arguments_from_model: '" ^ UnparseC.term_in_ctxt ctxt d ^ "' not in itms:\n" ^
  "itms:\n" ^ I_Model.to_string @{context} itms)
(* turn Model-items into arguments for a program *)
fun arguments_from_model ctxt mI itms =
  let
    val mvat = I_Model.max_variant itms
    fun okv mvat (_, vats, b, _, _) = member op = vats mvat andalso b
    val itms = filter (okv mvat) itms
    fun test_dsc d (_, _, _, _, itm_) = (d = I_Model.descriptor itm_)
    fun itm2arg itms (_, (d, _)) =
      case find_first (test_dsc d) itms of
        NONE => raise ERROR (error_msg_2 ctxt d itms)
      | SOME (_, _, _, _, itm_) => I_Model.penvval_in itm_
    val pats = (#model o MethodC.from_store ctxt) mI
    val _ = if pats = [] then raise ERROR error_msg_1 else ()
  in (flat o (map (itm2arg itms))) pats end;

(*
  convert a Prog_Tac to Tactic.input;
  argument Ctree.state is specifically for "Prog_Tac.SubProblem" (requires Pos),
*)
fun tac_from_prog (pt, pos) intac =
  let
    val pos = Pos.back_from_new pos
    val ctxt = Ctree.get_ctxt pt pos
    val thy = Proof_Context.theory_of ctxt
  in
    case intac of
      (Const (\<^const_name>\<open>Prog_Tac.Rewrite\<close>, _) $ thmID $ _) =>
        let
          val tid = HOLogic.dest_string thmID
        in (Tactic.Rewrite (tid, ThmC.thm_from_thy thy tid)) end
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Inst\<close>, _) $ sub $ thmID $ _) =>
        let
          val tid = HOLogic.dest_string thmID
          val sub' = Subst.program_to_input ctxt sub
        in (Tactic.Rewrite_Inst (sub', (tid, ThmC.thm_from_thy thy tid))) end
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set\<close>,_) $ rls $ _) =>
         (Tactic.Rewrite_Set (HOLogic.dest_string rls))
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set_Inst\<close>, _) $ sub $ rls $ _) =>
        let
          val sub' = Subst.program_to_input ctxt sub
        in (Tactic.Rewrite_Set_Inst (sub', HOLogic.dest_string rls)) end
      | (Const (\<^const_name>\<open>Prog_Tac.Calculate\<close>, _) $ op_ $ _) =>
          (Tactic.Calculate (HOLogic.dest_string op_))
      | (Const (\<^const_name>\<open>Prog_Tac.Take\<close>, _) $ t) =>
           (Tactic.Take (UnparseC.term_in_ctxt ctxt t))
      | (Const (\<^const_name>\<open>Prog_Tac.Substitute\<close>, _) $ isasub $ _) =>
          Tactic.Substitute (Prog_Tac.Substitute_adapt_to_type thy isasub)
      | (Const (\<^const_name>\<open>Prog_Tac.Check_elementwise\<close>, _) $ _ $ 
        (Const (\<^const_name>\<open>Collect\<close>, _) $ Abs (_, _, pred))) =>
          (Tactic.Check_elementwise (UnparseC.term_in_ctxt ctxt pred))
      | (Const (\<^const_name>\<open>Prog_Tac.Or_to_List\<close>, _) $ _ ) => Tactic.Or_to_List
      | (ptac as Const (\<^const_name>\<open>Prog_Tac.SubProblem\<close>, _) $ _ $ _) =>
          fst (Sub_Problem.tac_from_prog (pt, pos) ptac) (*might involve problem refinement etc*)
      | t =>
          raise ERROR ("stac2tac_ TODO: no match for " ^ UnparseC.term_in_ctxt ctxt t)
  end

datatype ass =
    Associated of
      Tactic.T         (* Subproblem' gets args instantiated in associate *)
      * term           (* resulting from application of Tactic.T          *)
      * Proof.context  (* updated by assumptioons from Rewrite*           *)
  | Ass_Weak of Tactic.T * term * Proof.context
  | Not_Associated;

(*
  associate is the ONLY code within by_tactic, which handles Tactic.T individually;
  thus it does ContextC.insert_assumptions in case of Rewrite*.
  The argument Ctree.ctree is required only for Subproblem'.
*)
fun trace_msg_3 ctxt tac =
  if Config.get ctxt LI_trace then
    tracing("@@@ the \"tac_\" proposed to apply does NOT match the leaf found in the program:\n" ^
      "@@@ tac_ = \"" ^ Tactic.string_of tac ^ "\"")
  else ();
fun associate _ ctxt (m as Tactic.Rewrite_Inst'
        (_, _, _, _, _, thm'' as (thmID, _), f, (f', asms')), stac) =
      (case stac of
  	    (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Inst\<close>, _) $ _ $ thmID_ $ f_) =>
  	      if thmID = HOLogic.dest_string thmID_ then
  	        if f = f_ then 
              Associated (m, f', ContextC.insert_assumptions asms' ctxt)
  	        else Ass_Weak (m, f', ContextC.insert_assumptions asms' ctxt)
  	      else Not_Associated
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set_Inst\<close>,_) $ _ $ rls_ $ f_) =>
  	      if Rule_Set.contains (Rule.Thm thm'') (get_rls ctxt (HOLogic.dest_string rls_)) then
            if f = f_ then
              Associated (m, f', ContextC.insert_assumptions asms' ctxt)
            else Ass_Weak (m, f', ContextC.insert_assumptions asms' ctxt)
  	      else Not_Associated
      | _ => Not_Associated)
  | associate _ ctxt (m as Tactic.Rewrite' (_, _, _, _, thm'' as (thmID, _), f, (f', asms')), stac) =
      (case stac of
  	    (Const (\<^const_name>\<open>Prog_Tac.Rewrite\<close>, _) $ thmID_ $ f_) =>
  	      if thmID = HOLogic.dest_string thmID_ then
  	        if f = f_ then
              Associated (m, f', ContextC.insert_assumptions asms' ctxt)
  	        else Ass_Weak (m, f', ContextC.insert_assumptions asms' ctxt)
  	      else Not_Associated
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set\<close>, _) $ rls_ $ f_) =>
  	       if Rule_Set.contains (Rule.Thm thm'') (get_rls ctxt (HOLogic.dest_string rls_)) then
             if f = f_ then
               Associated (m, f', ContextC.insert_assumptions asms' ctxt)
  	         else Ass_Weak (m, f', ContextC.insert_assumptions asms' ctxt)
  	       else Not_Associated
      | _ => Not_Associated)
  | associate _ ctxt (m as Tactic.Rewrite_Set_Inst' (_, _, _, rls, f, (f', asms')),
        (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set_Inst\<close>, _) $ _ $ rls_ $ f_)) = 
      if Rule_Set.id rls = HOLogic.dest_string rls_ then
        if f = f_ then
          Associated (m, f', ContextC.insert_assumptions asms' ctxt)
        else Ass_Weak (m ,f', ContextC.insert_assumptions asms' ctxt)
      else Not_Associated
  | associate _ ctxt (m as Tactic.Rewrite_Set' (_, _, rls, f, (f', asms')),
        (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set\<close>, _) $ rls_ $ f_)) = 
      if Rule_Set.id rls = HOLogic.dest_string rls_ then
        if f = f_ then
          Associated (m, f', ContextC.insert_assumptions asms' ctxt)
        else Ass_Weak (m ,f', ContextC.insert_assumptions asms' ctxt)
      else Not_Associated
  | associate _ ctxt (m as Tactic.Calculate' (_, op_, f, (f', _)), stac) =
      (case stac of
  	    (Const (\<^const_name>\<open>Prog_Tac.Calculate\<close>,_) $ op__ $ f_) =>
  	      if op_ = HOLogic.dest_string op__ then
            if f = f_ then Associated (m, f', ctxt) else Ass_Weak (m ,f', ctxt)
  	      else Not_Associated
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set_Inst\<close>, _) $ _ $ rls_ $ f_)  =>
          if Rule_Set.contains (Rule.Eval (get_calc ctxt
            op_ |> snd)) (get_rls ctxt (HOLogic.dest_string rls_)) then
            if f = f_ then Associated (m, f', ctxt) else Ass_Weak (m ,f', ctxt)
          else Not_Associated
      | (Const (\<^const_name>\<open>Prog_Tac.Rewrite_Set\<close>,_) $ rls_ $ f_) =>
          if Rule_Set.contains (Rule.Eval (get_calc ctxt
            op_ |> snd)) (get_rls ctxt (HOLogic.dest_string rls_)) then
            if f = f_ then Associated (m, f', ctxt) else Ass_Weak (m ,f', ctxt)
          else Not_Associated
      | _ => Not_Associated)
  | associate _ ctxt (m as Tactic.Check_elementwise' (consts, _, (consts_chkd, _)),
        (Const (\<^const_name>\<open>Prog_Tac.Check_elementwise\<close>,_) $ consts' $ _)) =
      if consts = consts' then Associated (m, consts_chkd, ctxt) else Not_Associated
  | associate _ ctxt (m as Tactic.Or_to_List' (_, list), (Const (\<^const_name>\<open>Prog_Tac.Or_to_List\<close>, _) $ _)) =
      Associated (m, list, ctxt)
  | associate _ ctxt (m as Tactic.Take' term, (Const (\<^const_name>\<open>Prog_Tac.Take\<close>, _) $ _)) =
      Associated (m, term, ctxt)
  | associate _ ctxt (m as Tactic.Substitute' (ro, asm_rls, subte, f, f'),
        (Const (\<^const_name>\<open>Prog_Tac.Substitute\<close>, _) $ _ $ t)) =
  	  if f = t then
  	   Associated (m, f', ctxt)
  	  else (*compare | applicable_in (p,p_) pt (m as Substitute sube)*)
  		  if foldl and_ (true, map TermC.contains_Var subte) then
  		    let
  		      val t' = subst_atomic (map HOLogic.dest_eq subte) t
  		    in
  		      if t = t' then raise ERROR "associate: Substitute' not applicable to val of Expr"
  		      else Associated (Tactic.Substitute' (ro, asm_rls, subte, t, t'), t', ctxt)
  		    end
  		  else
  		    (case Rewrite.rewrite_terms_ (Proof_Context.init_global (ThyC.Isac ())) ro asm_rls subte t of
  		      SOME (t', _) =>  Associated (Tactic.Substitute' (ro, asm_rls, subte, t, t'), t', ctxt)
  		    | NONE => raise ERROR "associate: Substitute' not applicable to val of Expr")
  | associate pt ctxt (Tactic.Subproblem' ((domID, pblID, _), _, _, _, _, _),
        (stac as Const (\<^const_name>\<open>Prog_Tac.SubProblem\<close>, _) $ _ $ _)) =
      (case Sub_Problem.tac_from_prog pt stac of
        (_, tac as Tactic.Subproblem' ((dI, pI, _), _, _, _, _, f)) =>
          if domID = dI andalso pblID = pI then Associated (tac, f, ctxt) else Not_Associated
      | _ => raise ERROR ("associate: uncovered case"))
  | associate _ ctxt (tac, _) = 
    (trace_msg_3 ctxt tac; Not_Associated);

(* in detail step find the next parent, which is either a PblObj xor a PrfObj *)
fun parent_node _ ([], _) = (true, [], Rule_Set.Empty, ContextC.empty)
  | parent_node pt (pos as (p, _)) =
    let
      fun par _ [] = (true, [], Rule_Set.Empty, ContextC.empty)
        | par pt p =
            if Ctree.is_pblobj (Ctree.get_obj I pt p)
            then (true, p, Rule_Set.Empty, ContextC.empty)
  		      else 
              let
                val ctxt = Ctree.get_ctxt pt pos
              in
                case Ctree.get_obj Ctree.g_tac pt p of
      				    Tactic.Rewrite_Set rls' => (false, p, get_rls ctxt rls', ctxt)
      			    | Tactic.Rewrite_Set_Inst (_, rls') => (false, p, get_rls ctxt rls', ctxt)
      			    | _ => par pt (Pos.lev_up p)
              end
    in par pt (Pos.lev_up p) end; 

(* create the initial interpreter state
  from the items of the guard and the formal arguments of the program.
Note on progression from (a) type fmz_ \<longrightarrow> (e) arguments of the partial_function:
  (a) fmz is given by a math author
  (b) fmz_ is transformed to ori list as a prerequisite for efficient interactive modelling
  (c) modelling creates an itm list from ori list + possible user input
  (d) specifying a theory might add some items and create a guard for the program
  (e) fun relate_args creates an environment for evaluating the program.
Note on sequence-relation (a) -- (e), i.e. the (formal) arguments of the program:
  * Since the arguments of the partial_function have no description (see Input_Descript.thy),
    (e) depends on the sequence in fmz_ and on the types of the formal arguments.
  * pairing formal arguments with itm's follows the sequence
  * Thus the resulting sequence-relation can be ambiguous.
  * Ambiguities are done by rearranging fmz_ or the formal arguments.
  * The latter is easier, albeit not optimal: a fmz_ can be used by different programs.
  *)
local
val errmsg = "ERROR: found no actual arguments for prog. of "
fun msg_miss ctxt sc metID caller f formals actuals =
  "ERROR in creating the environment from formal args. of partial_function \"" ^ fst (Program.get_fun_id sc) ^ "\"\n" ^
	"and the actual args., ie. items of the guard of \"" ^ MethodC.id_to_string metID ^ "\" by \"" ^ caller ^ "\":\n" ^
	"formal arg \"" ^ UnparseC.term_in_ctxt ctxt f ^ "\" doesn't match any actual arg.\n" ^
	"with:\n" ^
	(string_of_int o length) formals ^ " formal args: " ^ UnparseC.terms_in_ctxt ctxt formals ^ "\n" ^
	(string_of_int o length) actuals ^ " actual args: " ^ UnparseC.terms_in_ctxt ctxt actuals
fun msg_miss_type ctxt sc metID f formals actuals =
  "ERROR in creating the environment from formal args. of partial_function \"" ^ fst (Program.get_fun_id sc) ^ "\"\n" ^
	"and the actual args., ie. items of the guard of \"" ^ MethodC.id_to_string metID ^ "\" by \"assoc_by_type\":\n" ^
	"formal arg \"" ^ UnparseC.term_in_ctxt ctxt f ^ "\" of type \"" ^ UnparseC.typ_in_ctxt ctxt (type_of f) ^
  "\" doesn't mach any actual arg.\nwith:\n" ^
	(string_of_int o length) formals ^ " formal args: " ^ UnparseC.terms_in_ctxt ctxt formals ^ "\n" ^
	(string_of_int o length) actuals ^ " actual args: " ^ UnparseC.terms_in_ctxt ctxt actuals ^ "\n" ^
  "   with types: " ^ (strs2str o (map ((UnparseC.typ_in_ctxt ctxt) o type_of))) actuals
fun msg_ambiguous ctxt sc metID f aas formals actuals =
  "AMBIGUITY in creating the environment from formal args. of partial_function \"" ^ fst (Program.get_fun_id sc) ^ "\"\n" ^
	"and the actual args., ie. items of the guard of \"" ^ MethodC.id_to_string metID ^ "\" by \"assoc_by_type\":\n" ^
  "formal arg. \"" ^ UnparseC.term_in_ctxt ctxt f ^ "\" type-matches with several..."  ^
  "actual args. \"" ^ UnparseC.terms_in_ctxt ctxt aas ^ "\"\n" ^
  "selected \"" ^ UnparseC.term_in_ctxt ctxt (hd aas) ^ "\"\n" ^
	"with\n" ^
	"formals: " ^ UnparseC.terms_in_ctxt ctxt formals ^ "\n" ^
	"actuals: " ^ UnparseC.terms_in_ctxt ctxt actuals
fun trace_init ctxt metID =
  if Config.get ctxt LI_trace
  then tracing("@@@ program \"" ^ strs2str metID ^ "\"")
  else ();
in
fun init_pstate prog_rls ctxt itms metID =
  let
    val actuals = arguments_from_model ctxt metID itms
    val _ = if actuals <> [] then () else raise ERROR (errmsg ^ strs2str' metID)
    val (program, sc) = (case (#program o MethodC.from_store ctxt) metID of
           program as Rule.Prog sc => (trace_init ctxt metID; (program, sc))
       | _ => raise ERROR ("init_pstate with " ^ MethodC.id_to_string metID))
    val formals = Program.formal_args sc    
    fun assoc_by_type ctxt f aa =
      case filter (curry (fn (f, a) => type_of f = type_of a) f) aa of
        [] => raise ERROR (msg_miss_type ctxt sc metID f formals actuals)
      | [a] => (f, a)
      | aas as (a :: _) => (writeln (msg_ambiguous ctxt sc metID f aas formals actuals); (f, a));
	  fun relate_args _ (f::_)  [] = raise ERROR (msg_miss ctxt sc metID "relate_args" f formals actuals)
	    | relate_args env [] _ = env (*may drop Find?*)
	    | relate_args env (f::ff) (aas as (a::aa)) = 
	      if type_of f = type_of a 
	      then relate_args (env @ [(f, a)]) ff aa
        else
          let val (f, a) = assoc_by_type ctxt f aas
          in relate_args (env @ [(f, a)]) ff (remove op = a aas) end
    val {where_, where_rls, ...} = MethodC.from_store ctxt metID;
    val pres = Pre_Conds.check ctxt where_rls where_ itms 0 |> snd |> map snd;
    val ctxt = ctxt |> ContextC.insert_assumptions pres;
    val ist = Istate.e_pstate
      |> Istate.set_eval prog_rls
      |> Istate.set_env_true (relate_args [] formals actuals)
  in
    (Istate.Pstate ist, ctxt, program)
  end;
end (*local*)

(* get the simplifier of the current method *)
fun get_simplifier (pt, pos as (p, _)) =
  let
    val p' = Ctree.par_pblobj pt p
	  val metID = Ctree.get_obj Ctree.g_metID pt p'
	  val ctxt = Ctree.get_ctxt pt pos
	  val {prog_rls, ...} = MethodC.from_store ctxt metID
  in prog_rls end

(* resume program interpretation at the beginning of a step *)
fun resume_prog pos pt =
  let
    val (is_problem, p', rls', ctxt) = parent_node pt pos (*is Ctree.PrfObj in case of detail step*)
  in
    if is_problem then
      let
	      val metID = get_obj g_metID pt p'
	      val {prog_rls, ...} = MethodC.from_store ctxt metID
	      val (is, ctxt) =
	        case get_loc pt pos of
	           (Istate.Pstate ist, ctxt) => (Istate.Pstate (Istate.set_eval prog_rls ist), ctxt)
	        | _ => raise ERROR "resume_prog: unexpected result from get_loc"
	    in
	      ((is, ctxt), (#program o MethodC.from_store ctxt) metID)
	    end
    else
      (get_loc pt pos,
      Rule.Prog (Auto_Prog.gen (Proof_Context.theory_of ctxt) (get_last_formula (pt, pos)) rls'))
  end

fun trace_msg_1 ctxt call t stac =
  if Config.get ctxt LI_trace then
	  tracing ("@@@ " ^ call ^ " leaf \"" ^ UnparseC.term_in_ctxt ctxt t ^ "\" \<longrightarrow> Tac \"" ^
	    UnparseC.term_in_ctxt ctxt stac ^ "\"")
	else ();
fun trace_msg_2 ctxt call t lexpr' =
  if Config.get ctxt LI_trace then
	  tracing("@@@ " ^ call ^ " leaf '" ^ UnparseC.term_in_ctxt ctxt t ^ "' \<longrightarrow> Expr \"" ^
	    UnparseC.term_in_ctxt ctxt lexpr' ^ "\"")
	else ();
(*
  check a leaf at the end of recursive descent; a leaf is either a Prog_Tac or a Prog_Expr.
  snd of return value are the formal arguments in case of currying.
*)
fun check_leaf call ctxt prog_rls (E, (a, v)) t =
    case Prog_Tac.eval_leaf E a v t of
	    (Program.Tac stac, a') =>
	      let
	        val stac' = Rewrite.eval_prog_expr ctxt prog_rls 
              (subst_atomic (Env.update_opt E (a, v)) stac)
	      in
          (trace_msg_1 ctxt call t stac; (Program.Tac stac', a'))
	      end
    | (Program.Expr lexpr, a') =>
	      let
	        val lexpr' = Rewrite.eval_prog_expr ctxt prog_rls
              (subst_atomic (Env.update_opt E (a, v)) lexpr)
	      in
          (trace_msg_2 ctxt call t lexpr'; (Program.Expr lexpr', a'))
	      end;

(**)end(**)
